/*
USAGE after compiling:
./false <filename> [input]
like ./false cat.false "Hello World"

Exit codes:
0 - success
1 - No file specified
2 - Memory error (Couldn't realloc(...))
3 - Unknown tokentype
4 - Invoked undefined behaviour
5 - Coudln't open specified file
*/

#include <stdio.h> 
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include "tokenstack.h"
#include "stringlib.h"
#include "utils.h"
#include "token.h"

#define current_char checkChar(string_cstr(instructions)[current])

typedef struct program{
    stack_t* stck;
    token_t** variables;
    string_t* input;
    bool* initialized_vars;
} program_t;


char checkChar(char c){
    if(!((9 <= c && c <= 10) || (32 <= c && c <= 126))){
        printf("\nUNDEFINED BEHAVIOUR - char '%c' out of valid range", c);
        exit(4);
    }
    return c;
}

char eat_char_from_input(string_t* input){
    if(string_get_size(input) == 0){
        printf("\nUNDEFINED BEHAVIOUR - Trying to get char from input while being empty - Exiting...\n");
        exit(4);
    }
    const char inp = string_cstr(input)[0];
    string_erase(input, 0, 1);
    return checkChar(inp);
}

void run_quote(string_t* instructions, program_t* context){
    unsigned long long current = 0;
    const unsigned long long instructions_size = string_get_size(instructions);
    string_t* buffer = string_new();
    while(current < instructions_size){
        switch(current_char){
            case '{':
                {
                    char symbol = string_cstr(instructions)[current];
                    while(symbol != '}' && ++current < instructions_size){
                        symbol = current_char;
                    }
                    if(current==instructions_size){
                        printf("UNDEFINED BEHAVIOUR - Comment was never closed - Exiting...\n");
                        exit(4);
                    }
                    ++current;
                }
                break;

            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                {
                    char symbol = current_char;
                    do
                    {
                        string_append_char(buffer, symbol);
                        if(++current >= instructions_size){
                            break;
                        }
                        symbol = current_char;
                    } while ('0' <= symbol && symbol <= '9');
                    stack_push(context->stck, token_new_int(strtol(string_cstr(buffer), NULL, 10)));
                    string_clear(buffer);
                }
                break;

            case '\'':
                ++current;
                stack_push(context->stck, token_new_int(current_char));
                ++current;
                break;
            
            case '"':
                {
                    ++current;
                    if (current >= instructions_size){
                        printf("\nUNDEFINED BEHAVIOUR - String never closed - Exiting...\n");
                        exit(4);
                    }
                    
                    char symbol = current_char;
                    do{
                        string_append_char(buffer, symbol);
                        if (++current >= instructions_size){
                            printf("\nUNDEFINED BEHAVIOUR - Quote never closed - Exiting...\n");
                            exit(4);
                        }
                        symbol = current_char;
                    } while (symbol != '"');
                    printf("%s", string_cstr(buffer));
                    string_clear(buffer);
                    ++current;

                }
                break;

            case '[':
                {
                    ++current;
                    if (current >= instructions_size){
                        printf("\nUNDEFINED BEHAVIOUR - Quote never closed - Exiting...\n");
                        exit(4);
                    }
                    char symbol = current_char;
                    unsigned long long open_bracket_count = current_char != '[' ? 1 : 2;
                    do
                    {
                        string_append_char(buffer, symbol);
                        if (++current >= instructions_size){
                            printf("\nUNDEFINED BEHAVIOUR - Quote never closed - Exiting...\n");
                            exit(4);
                        }
                        symbol = current_char;
                        if(symbol == '['){
                            ++open_bracket_count;
                        }
                        else if(symbol == ']'){
                            --open_bracket_count;
                        }


                    } while (symbol != ']' || open_bracket_count > 0);
                    stack_push(context->stck, token_new_quote(string_copy(buffer)));
                    string_clear(buffer);
                    ++current;
                }
                break;
            
            case 'a':
            case 'b':
            case 'c':
            case 'd':
            case 'e':
            case 'f':
            case 'g':
            case 'h':
            case 'i':
            case 'j':
            case 'k':
            case 'l':
            case 'm':
            case 'n':
            case 'o':
            case 'p':
            case 'q':
            case 'r':
            case 's':
            case 't':
            case 'u':
            case 'v':
            case 'w':
            case 'x':
            case 'y':
            case 'z':
                stack_push(context->stck, token_new_reference(current_char));
                ++current;
                break;

            case ';':
                {
                    token_t* reference = stack_pop(context->stck);
                    if(reference->type != REFERENCE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Reference - Exiting...\n");
                        exit(4);
                    }
                    if(!context->initialized_vars[reference->value_r-'a']){
                        printf("\nUNDEFINED BEHAVIOUR - Reference '%c' is not initialized - Exiting...\n", reference->value_r);
                        exit(4);
                    }
                    token_t* new_tok = token_copy(context->variables[reference->value_r-'a']);
                    stack_push(context->stck, new_tok);
                    token_free(reference);
                    ++current;
                }
                break;

            case ':':
                {
                    token_t* reference = stack_pop(context->stck);
                    if(reference->type != REFERENCE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Reference - Exiting...\n");
                        exit(4);
                    }
                    if(context->initialized_vars[reference->value_r-'a']){
                        token_full_free(context->variables[reference->value_r-'a']);
                    }
                    context->variables[reference->value_r-'a'] = stack_pop(context->stck);
                    context->initialized_vars[reference->value_r-'a'] = true;
                    token_free(reference);
                    current++;
                }
                break;

            case '$':
                stack_push(context->stck, token_copy(stack_raccess(context->stck, 0)));
                ++current;
                break;

            case '%':
                token_full_free(stack_pop(context->stck));
                ++current;
                break;
            
            case '\\':
                {
                    token_t* first = stack_pop(context->stck);
                    token_t* second = stack_pop(context->stck);
                    stack_push(context->stck, first);
                    stack_push(context->stck, second);
                    ++current;
                }
                break;
            
            case '@':
                {
                    token_t* first = stack_pop(context->stck);
                    token_t* second = stack_pop(context->stck);
                    token_t* third = stack_pop(context->stck);
                    stack_push(context->stck, second);
                    stack_push(context->stck, first);
                    stack_push(context->stck, third);
                    ++current;
                }
                break;
            
            case 'O':
                {
                    token_t* index = stack_pop(context->stck);
                    if(index->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_copy(stack_raccess(context->stck, index->value_i)));
                    token_free(index);
                    ++current;
                }
                break;
            
            case '+':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(first->value_i + second->value_i));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;
            
            case '-':
                {
                    token_t* subtrahend = stack_pop(context->stck);
                    if(subtrahend->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* minuend = stack_pop(context->stck);
                    if(minuend->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(minuend->value_i - subtrahend->value_i));
                    token_free(subtrahend);
                    token_free(minuend);
                    ++current;
                }
                break;

            case '*':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(first->value_i * second->value_i));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;
            
            case '/':
                {
                    token_t* divisor = stack_pop(context->stck);
                    if(divisor->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* dividend = stack_pop(context->stck);
                    if(dividend->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(dividend->value_i / divisor->value_i));
                    token_free(divisor);
                    token_free(dividend);
                    ++current;
                }
                break;
            
            case '_':
                {
                    token_t* value = stack_pop(context->stck);
                    if(value->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(-(value->value_i)));
                    token_free(value);
                    ++current;
                }
                break;
            
            case '&':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(first->value_i & second->value_i));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;
            
            case '|':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(first->value_i | second->value_i));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;

            case '~':
                {
                    token_t* value = stack_pop(context->stck);
                    if(value->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(~(value->value_i)));
                    token_free(value);
                    ++current;
                }
                break;

            case '=':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(first->value_i == second->value_i ? -1 : 0));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;
            
            case '>':
                {
                    token_t* first = stack_pop(context->stck);
                    if(first->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    token_t* second = stack_pop(context->stck);
                    if(second->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    stack_push(context->stck, token_new_int(second->value_i > first->value_i ? -1 : 0));
                    token_free(first);
                    token_free(second);
                    ++current;
                }
                break;

            case '!':
                {
                    token_t* quote = stack_pop(context->stck);
                    if(quote->type != QUOTE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Quote - Exiting...\n");
                        exit(4);
                    }
                    run_quote(quote->value_q, context);
                    token_full_free(quote);
                    ++current;
                }
                break;
            
            case '?':
                {
                    token_t* quote = stack_pop(context->stck);
                    if(quote->type != QUOTE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Quote - Exiting...\n");
                        exit(4);
                    }
                    token_t* condition = stack_pop(context->stck);
                    if(condition->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    if(condition->value_i != 0){
                        run_quote(quote->value_q, context);
                    }
                    token_full_free(quote);
                    token_free(condition);
                    ++current;
                }
                break;
            
            case '#':
                {
                    token_t* body = stack_pop(context->stck);
                    if(body->type != QUOTE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Quote - Exiting...\n");
                        exit(4);
                    }
                    token_t* condition = stack_pop(context->stck);
                    if(condition->type != QUOTE){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not a Quote - Exiting...\n");
                        exit(4);
                    }
                    while(true){
                        run_quote(condition->value_q, context);
                        token_t* evaluated_condition = stack_pop(context->stck);
                        if(evaluated_condition->type != INTEGER){
                            printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                            exit(4);
                        }
                        if(evaluated_condition->value_i == 0){
                            token_free(evaluated_condition);
                            break;
                        }
                        token_free(evaluated_condition);
                        run_quote(body->value_q, context);
                    }
                    token_full_free(body);
                    token_full_free(condition);

                    ++current;
                }
                break;
            
            case '^':
                if(string_get_size(context->input)==0){
                    stack_push(context->stck, token_new_int(-1));
                }
                else{
                    stack_push(context->stck, token_new_int(eat_char_from_input(context->input)));
                }
                ++current;
                break;
            
            case ',':
                {
                    token_t* c = stack_pop(context->stck);
                    if(c->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    printf("%c", c->value_i);
                    token_free(c);
                    ++current;
                }
                break;

            case '.':
                {
                    token_t* c = stack_pop(context->stck);
                    if(c->type != INTEGER){
                        printf("\nUNDEFINED BEHAVIOUR - Pop'd object is not an Integer - Exiting...\n");
                        exit(4);
                    }
                    printf("%d", c->value_i);
                    token_free(c);
                    ++current;
                }
                break;

            case 'B':
                fflush(stdout);
                ++current;
                break;
            
            case '`':
                printf("\nUNDEFINED BEHAVIOUR - ` only does undefined behaviour - Exiting... \n");
                exit(4);
                break;

            default:
                switch(current_char){
                    case ' ':
                    case '\t':
                    case '\n':
                        ++current;
                        break;
                    default:
                        printf("\nUNDEFINED BEHAVIOUR - Unknown instruction '%c' - Exiting...\n", current_char);
                        exit(4);
                }
        }
    }
    string_free(buffer);
    return;
}

int main(int argc, char const *argv[])
{
    if(argc == 1){
        printf("No file specified\n");
        exit(1);
    }
    const char* filename = argv[1];    
    string_t* input = string_new();
    if(argc > 2){
        string_append_chars(input, argv[2]);
    }

    FILE* file = fopen(filename, "r");
    if(file == NULL){
        string_free(input);
        printf("Couldn't open file \"%s\"\n", filename);
        exit(5);
    }

    fseek(file, 0, SEEK_END);
    unsigned long long filesize = ftell(file);
    rewind(file);
    char* filecontent = malloc((filesize+1)*sizeof(char));
    fread(filecontent, filesize, 1, file); 
    filecontent[filesize] = '\0';
    string_t* instructions = string_from_chars(filecontent);
    free(filecontent);
    fclose(file);

    stack_t* stck = stack_new();
    token_t* variables[26];
    bool initialized_vars[26] = {0};
    
    program_t* context = malloc(sizeof(program_t));
    context->stck = stck;
    context->variables = variables;
    context->input = input;
    context->initialized_vars = initialized_vars;

    run_quote(instructions, context);

    string_free(instructions);
    stack_free(stck);
    string_free(input);
    for(int i = 0; i < 26; ++i){
        if(initialized_vars[i]){
            token_full_free(variables[i]);
        }
    }
    free(context);
}
