#include <stdlib.h>
#include <stdio.h>
#include "token.h"
#include "tokenstack.h"
#include "utils.h"

typedef struct stack{
    token_t** tokens;
    unsigned long long size;
    unsigned long long capacity;
} stack_t;

stack_t* stack_new(){
    stack_t* new_stack = malloc(sizeof(stack_t));
    new_stack->tokens = malloc(MIN_STACK_CAPACITY*sizeof(token_t));
    new_stack->size = 0;
    new_stack->capacity = MIN_STACK_CAPACITY;
    return new_stack;
}

token_t** stack_get_raw(const stack_t* stck){
    return stck->tokens;
}

unsigned long long stack_get_size(const stack_t* stck){
    return stck->size;
}

unsigned long long stack_get_capacity(const stack_t* stck){
    return stck->capacity;
}

void stack_set_min_capacity(stack_t* stck, const unsigned long long size){
    if(size > stck->capacity){
        token_t** new_tokens = realloc(stck->tokens, size*(sizeof(size_t*)));
        if(new_tokens==NULL){
            exit(2);
        }
        stck->tokens = new_tokens;
        stck->capacity = size;
    }
    return;
}

void stack_push(stack_t* stck, token_t* tok){
    const unsigned long long size = stck->size;
    stack_set_min_capacity(stck, umax(next_power_of_2(size+1), MIN_STACK_CAPACITY));
    stck->tokens[size] = tok;
    stck->size = size+1;
    return;
}

token_t* stack_pop(stack_t* stck){
    if(stck->size == 0){
        printf("\nUNDEFINED BEHAVIOUR - Stack has been pop'd while being empty - Exiting...\n");
        exit(4);
    }
    stck->size = stck->size-1;
    return stck->tokens[stck->size];
}

token_t* stack_access(stack_t* stck, const unsigned long long index){
    if(index >= stck->size){
        printf("\nUNDEFINED BEHAVIOUR - Stack access out of bounds (stacksize: %llu, index: %llu) - Exiting...\n", stck->size, index);
        exit(4);
    }
    return stck->tokens[index];
}

token_t* stack_raccess(stack_t* stck, const unsigned long long index){
    if(index >= stck->size){
        printf("\nUNDEFINED BEHAVIOUR - Stack raccess out of bounds (stacksize: %llu, index: %llu) - Exiting...\n", stck->size, index);
        exit(4);
    }
    return stck->tokens[stck->size - 1 - index];
}

void stack_print(const stack_t* stck){
    printf("[");
    for(unsigned long long i = 0; i < stck->size; ++i){
        token_print(stck->tokens[i]);
        printf(", ");
    }
    printf("]");
    return;
}

void stack_free(stack_t* stck){
    for(unsigned long long i = 0; i < stck->size; ++i){
        token_full_free(stck->tokens[i]);
    }
    free(stck->tokens);
    free(stck);
    return;
}

